/*
 The software in this package is distributed under the GNU General
 Public License version 2 (with a special exception described below).

 A copy of GNU General Public License (GPL) is included in this distribution,
 in the file COPYING.GPL.

 As a special exception, if other files instantiate templates or use macros
 or inline functions from this file, or you compile this file and link it
 with other works to produce a work based on this file, this file
 does not by itself cause the resulting work to be covered
 by the GNU General Public License.

 However the source code for this file must still be made available
 in accordance with section (3) of the GNU General Public License.

 This exception does not invalidate any other reasons why a work based
 on this file might be covered by the GNU General Public License.
 */

#ifndef NETWORK_INTERFACE_HPP
#define NETWORK_INTERFACE_HPP

#include <errno.h>
#include <string.h>

#include <arpa/inet.h>
#include <ifaddrs.h>
#include <netinet/in.h>
#include <net/if.h>
#include <sys/socket.h>

#include <string>

#include <boost/asio/basic_socket.hpp>
#include <boost/asio/basic_raw_socket.hpp>
#include <boost/asio/ip/address.hpp>
#include <boost/asio/ip/address_v4.hpp>
#include <boost/asio/ip/address_v6.hpp>

#include <logfunc.hpp>

#include "boost_assert_handler.h"
#include "host/networkinterfacelist.h"

typedef const struct sockaddr* const_sockaddr_ptr_t;
typedef const struct sockaddr_in* const_sockaddr_in_ptr_t;
typedef const struct sockaddr_in6* const_sockaddr_in6_ptr_t;

//-----------------------------------------------------------------------------
// NetworkInterface
//-----------------------------------------------------------------------------

/**
 * @brief This class represents a local network interface.
 */
template<typename SocketType, typename Protocol>
class NetworkInterface
{
public:

    NetworkInterface(
            const std::string &nic_name,
            SocketType &boost_socket
    );

    bool bind();

    std::string get_name() const;

    boost::asio::ip::address get_address( Protocol protocol ) const;

    const uint8_t* get_address_bytes( const_sockaddr_ptr_t ifa_addr ) const;

private:
    /// The network interface name which the socket is attached
    const std::string Name;
    /// The socket used in this network interface
    SocketType &Socket;

};


//-----------------------------------------------------------------------------
// NetworkInterface
//-----------------------------------------------------------------------------

template<typename SocketType, typename Protocol>
NetworkInterface<SocketType, Protocol>::NetworkInterface(
        const std::string &nic_name,
        SocketType &boost_socket
) :
    Name( nic_name ),
    Socket( boost_socket )
{
    BOOST_ASSERT( !nic_name.empty() );
}

/**
 * @brief Avoid the socket to drop to another network interface if the
 * destination is unreachable through the binded interface. Packets are
 * sent only from this interface.
 *
 * @return true if the bind was successfully, and false if the bind failed.
 */
template<typename SocketType, typename Protocol>
bool NetworkInterface<SocketType, Protocol>::bind()
{
    BOOST_ASSERT( !Name.empty() );

    int ret = ::setsockopt(
            Socket.native(),
            SOL_SOCKET,
            SO_BINDTODEVICE,
            Name.c_str(),
            Name.size()
    );
    if ( ret == -1 )
    {
        I2n::Logger::GlobalLogger.error()
            << "Could not bind pinger to interface " << Name << std::endl;
        return false;
    }

    return true;
}

/**
 * @brief Get the network interface name.
 *
 * @return A string representing the name of the network interface.
 */
template<typename SocketType, typename Protocol>
std::string NetworkInterface<SocketType, Protocol>::get_name() const
{
    BOOST_ASSERT( !Name.empty() );

    return Name;
}

/**
 * @brief This method gets the local address of network interface.
 *
 * @return The @c boost::asio::ip::address representing the local address of
 * the network interface.
 */
template<typename SocketType, typename Protocol>
boost::asio::ip::address NetworkInterface<SocketType, Protocol>::get_address(
        Protocol protocol
) const
{
    BOOST_ASSERT( (protocol.family() == PF_INET) || (protocol.family() == PF_INET6) );

    NetworkInterfaceList nic_list;

    const sa_family_t addr_type = static_cast<sa_family_t>( protocol.family() );
    const struct ifaddrs *ifa_entry = nic_list.find_interface( get_name(), addr_type );
    if ( ifa_entry == NULL )
    {
        return boost::asio::ip::address();
    }

    const uint8_t *addr_bytes = get_address_bytes( ifa_entry->ifa_addr );
    if ( addr_bytes != NULL )
    {
        BOOST_ASSERT( ifa_entry->ifa_addr->sa_family == addr_type );

        char addr_buffer_string[ INET6_ADDRSTRLEN ];
        volatile const char *addr_ret_string = inet_ntop( addr_type, addr_bytes, addr_buffer_string, sizeof(addr_buffer_string) );
        if ( addr_ret_string != NULL )
        {
            if ( PF_INET == addr_type )
            {
                return boost::asio::ip::address_v4::from_string( std::string( addr_buffer_string ) );
            }
            else if ( PF_INET6 == addr_type )
            {
                return boost::asio::ip::address_v6::from_string( std::string( addr_buffer_string ) );
            }
        }
    }

    return boost::asio::ip::address();
}

/**
 * @brief Converts @c sockaddr into a sequence of bytes.
 *
 * @param ifa_addr A @c sockaddr structure.
 *
 * @return A sequence of bytes representing the interface address.
 */
template<typename SocketType, typename Protocol>
const uint8_t* NetworkInterface<SocketType, Protocol>::get_address_bytes(
        const_sockaddr_ptr_t ifa_addr
) const
{
    BOOST_ASSERT( ifa_addr != NULL );

    sa_family_t addr_type = ifa_addr->sa_family;
    const uint8_t *addr_bytes = NULL;

    if ( PF_INET == addr_type )
    {
        const_sockaddr_in_ptr_t sockaddr = reinterpret_cast<const_sockaddr_in_ptr_t>( ifa_addr );
        addr_bytes = reinterpret_cast<const uint8_t*>( & (sockaddr->sin_addr) );
    }
    else if ( PF_INET6 == addr_type )
    {
        const_sockaddr_in6_ptr_t sockaddr = reinterpret_cast<const_sockaddr_in6_ptr_t>( ifa_addr );
        addr_bytes = reinterpret_cast<const uint8_t*>( & (sockaddr->sin6_addr) );
    }

    return addr_bytes;
}

#endif // NETWORK_INTERFACE_HPP
